'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _is = require('./is');

var _is2 = _interopRequireDefault(_is);

var _utils = require('./utils');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var same = function same(v) {
  return v;
};
var isNonEmptyString = function isNonEmptyString(s) {
  return _is2.default.String(s) && !s;
};

/**
 * session/local storage can set/get object/array
 */

function storageFactory(storageInstance) {
  var storage = {};
  var sringifyRegexp = /^(_array_|_object_)\|/;
  var type = storageInstance === sessionStorage ? 'session' : 'storage';

  storage.set = function (name, value) {
    var val = value;
    (0, _utils.invariant)(_is2.default.Array(value) || _is2.default.PlainObject(value) || _is2.default.String(value), 'Hope: Expecting value of ' + type + '.set is a plain-object or array or string in instanceof check, but got ' + (0, _is.kindOf)(value));
    if (_is2.default.Array(value)) {
      val = '_array_|' + JSON.stringify(value);
    } else if (_is2.default.PlainObject(value)) {
      val = '_object_|' + JSON.stringify(value);
    }
    storageInstance.setItem(name, val);
  };

  storage.get = function (name) {
    var value = storageInstance.getItem(name);
    try {
      if (sringifyRegexp.test(value)) {
        return JSON.parse(value.replace(sringifyRegexp, ''));
      }
    } catch (err) {
      (0, _utils.log)('error', 'Hope: A error has occurred when read the \'' + name + '\'.', err);
    }
    return value;
  };

  storage.has = function (name) {
    return _is2.default.Defined(storage.get(name));
  };

  storage.clear = storageInstance.clear.bind(storageInstance);
  storage.remove = storageInstance.removeItem.bind(storageInstance);

  storage.forEach = function (callback) {
    var context = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;

    (0, _utils.invariant)(_is2.default.Function(callback), 'Hope: Expecting callback of ' + type + '.forEach is a function in instanceof check, but got ' + (0, _is.kindOf)(callback));
    var i = 0;
    var l = storageInstance.length;
    for (; i < l; i++) {
      var key = storageInstance.key(i);
      callback.call(context, key, storage.get(key));
    }
  };

  storage.getAll = function () {
    var all = {};
    storage.forEach(function (key, val) {
      return all[key] = val;
    });
    return all;
  };

  return storage;
}

/**
 * browser cookie
 */

var checkCookieName = function checkCookieName(name) {
  (0, _utils.invariant)(isNonEmptyString(name), 'Hope: Cookie name must be a non-empty string.');
};

var decode = function decode(value) {
  try {
    return decodeURIComponent(value);
  } catch (err) {
    (0, _utils.log)('error', 'Hope: A error has occurred in decodeURIComponent.', err);
  }
  return value;
};

var encode = function encode(value) {
  try {
    return encodeURIComponent(value);
  } catch (err) {
    (0, _utils.log)('error', 'Hope: A error has occurred in encodeURIComponent.', err);
  }
  return value;
};

var parseCookie = function parseCookie(str) {
  var obj = {};
  var pairs = str.split(/\s*;\s*/);

  if (pairs[0] === '') return obj;

  pairs.forEach(function (pair, i) {
    var _p = pair[i].split('=');
    obj[decode(_p[0])] = decode(_p[1]);
  });

  return obj;
};

var cookieFactory = function cookieFactory() {
  var cookie = {};

  cookie.set = function (name, value) {
    var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    var expires = options.expires || options.maxage;
    var secure = options.secure;
    var domain = options.domain;
    var path = options.path;

    checkCookieName(name);

    (0, _utils.invariant)(_is2.default.Number(expires) || _is2.default.Undefined(expires), 'Hope: Cookie expires/maxage must be a number.');

    (0, _utils.invariant)(_is2.default.Boolean(secure) || _is2.default.Undefined(expires), 'Hope: Cookie secure must be a boolean.');

    (0, _utils.invariant)(_is2.default.String(domain) || _is2.default.Undefined(expires), 'Hope: Cookie domain must be a string.');

    (0, _utils.invariant)(_is2.default.String(path) || _is2.default.Undefined(expires), 'Hope: Cookie path must be a string.');

    if (value === null || value === '') {
      expires = -1;
    }

    var txt = encode(name) + '=' + encode(String(value));

    if (expires) {
      expires = new Date(+new Date() + expires);
      txt += '; expires=' + expires.toUTCString();
    }

    if (isNonEmptyString(domain)) txt += '; domain=' + domain;
    if (isNonEmptyString(path)) txt += '; path=' + path;
    if (secure) txt += '; secure';

    document.cookie = txt;
  };

  cookie.getAll = function () {
    try {
      return parseCookie(document.cookie);
    } catch (err) {
      (0, _utils.log)('error', 'Hope: A error has occurred in parseCookie.', err);
    }
    return {};
  };

  cookie.get = function (name) {
    var opts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    checkCookieName(name);
    var options = Object.create(opts);
    if (_is2.default.Function(opts)) {
      options = { converter: options };
    }
    return (options.converter || same)(cookie.getAll()[name]);
  };

  cookie.forEach = function (callback) {
    var all = cookie.getAll();
    Object.keys(all).forEach(callback);
    return cookie;
  };

  cookie.forEach = function (callback) {
    var all = cookie.getAll();
    Object.keys(all).forEach(callback);
    return cookie;
  };
  cookie.has = function (name) {
    return cookie.get(name) !== undefined;
  };
  cookie.clear = function () {
    return cookie.forEach(function (name) {
      return cookie.remove(name);
    });
  };
  cookie.remove = function (name, options) {
    return cookie.set(name, null, options);
  };

  return window && cookie;
};

var isNode = false;

try {
  window.document;
} catch (e) {
  isNode = true;
}

exports.default = {

  // sessionStorage
  session: isNode ? null : storageFactory(sessionStorage),

  // localStorage
  storage: isNode ? null : storageFactory(localStorage),

  cookie: isNode ? null : cookieFactory()
};
module.exports = exports['default'];